/*
 * Copyright 2007 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/queue.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <time.h>

#include <event.h>
#include <dnet.h>

#include "dnsscan.h"
#include "exclude.h"
#include "checkpoint.h"
#include "dnsdata.gen.h"

static int
fill_in_generators(struct State *state, struct generators *generators)
{
	struct generate *gen;
	struct addr addr;

	if (EVTAG_ASSIGN(state, count, generators->count) == -1)
		goto error;

	TAILQ_FOREACH(gen, &generators->generators, next) {
		struct StateGenerate *sgen = EVTAG_ADD(state, generators);
		assert(sgen != NULL);

		addr.addr_type = ADDR_TYPE_IP;
		addr.addr_bits = gen->bits;
		addr.addr_ip = gen->prefix;

		if (EVTAG_ASSIGN(sgen, addr,
			(uint8_t *)&addr, sizeof(addr)) == -1)
			goto error;
		if (EVTAG_ASSIGN(sgen, last_count, gen->last_count) == -1)
			goto error;
	}

	return (0);

error:
	err(1, "Cannot marshal data");
	/* NOT REACHED */
}

static int
fill_in_requests(struct State *state, struct requestq *requests)
{
	struct request *req;

	TAILQ_FOREACH(req, requests, next) {
		struct Address *address = EVTAG_ADD(state, addresses);
		assert(address != NULL);

		if (EVTAG_ASSIGN(address, addr,
			(uint8_t *)&req->dest, sizeof(req->dest)) == -1)
			goto error;
	}

	return (0);

error:
	err(1, "Cannot marshal data");
	/* NOT REACHED */
}

static int
fill_in_interfaces(struct State *state, struct interfaceq *interfaces)
{
	extern struct requestq http_requests;
	struct interface *inf;

	TAILQ_FOREACH(inf, interfaces, next) {
		fill_in_requests(state, &inf->pending);
		fill_in_requests(state, &inf->inflight);
		fill_in_requests(state, &http_requests);
	}

	return (0);
}

int
create_checkpoint(const char *filename,
    struct generators *generators,
    struct interfaceq *interfaces)
{
	extern uint32_t rndsbox[];
	struct evbuffer *buf = NULL;
	static char tmpname[1024];
	struct State *state = NULL;
	int fd = -1;

	snprintf(tmpname, sizeof(tmpname), "%s.tmp", filename);
	fd = open(tmpname, O_WRONLY | O_CREAT | O_TRUNC, 0600);
	if (fd == -1) {
		warn("cannot create %s", tmpname);
		return (-1);
	}

	state = State_new();
	if (state == NULL) {
		warn("out of memory for check pointing");
		close(fd);
		unlink(tmpname);
		return (-1);
	}

	fill_in_generators(state, generators);
	fill_in_interfaces(state, interfaces);

	if (EVTAG_ASSIGN(state, rndsbox,
		(uint8_t *)rndsbox, RNDSBOXSIZE * sizeof(uint32_t)) == -1)
		goto error;

	/* marshal the data into an event buffer */

	buf = evbuffer_new();
	assert(buf != NULL);

	evtag_marshal_State(buf, CHECKPOINT_TAG, state);

	State_free(state);

	while (EVBUFFER_LENGTH(buf) > 0) {
		if (evbuffer_write(buf, fd) == -1) {
			if (errno != EAGAIN)
				goto error;
		}
	}

	evbuffer_free(buf);
	close(fd);

	if (rename(tmpname, filename) == -1) {
		warn("rename from %s to %s failed", tmpname, filename);
		return (-1);
	}

	return (0);

error:
	if (fd != -1)
		close(fd);
	if (buf != NULL)
		evbuffer_free(buf);

	return (-1);
}

static int
read_generators(struct State *state, struct generators *generators)
{
	struct generate *gen;
	struct addr *addr, net;
	int i;

	if (EVTAG_GET(state, count, &generators->count) == -1)
		return (-1);

	fprintf(stderr, "... current %u\n", generators->count);
	
	for (i = 0; i < EVTAG_LEN(state, generators); ++i) {
		struct StateGenerate *sg;
		uint8_t *data;
		uint32_t data_len;

		if (EVTAG_GET(state, generators, i, &sg) == -1)
			return (-1);

		gen = calloc(1, sizeof(struct generate));
		assert(gen != NULL);

		if (EVTAG_GET(sg, addr, &data, &data_len) == -1)
			return (-1);
		if (data_len != sizeof(struct addr))
			return (-1);

		addr = (struct addr *)data;
		addr_net(addr, &net);

		gen->prefix = net.addr_ip;
		gen->bits = addr->addr_bits;

		if (EVTAG_GET(sg, last_count, &gen->last_count) == -1)
			return (-1);

		DFPRINTF(1, (stderr, " ... read %s (%u)\n",
			addr_ntoa(addr), gen->last_count));

		TAILQ_INSERT_TAIL(&generators->generators, gen, next);
	}

	return (0);
}

static int
read_requests(struct State *state, struct interfaceq* interfaces)
{
	struct interface *inf = NULL;

	int i;
	for (i = 0; i < EVTAG_LEN(state, addresses); ++i) {
		struct Address *a;
		struct addr addr;
		uint8_t *data;
		uint32_t data_len;

		if (EVTAG_GET(state, addresses, i, &a) == -1)
			return (-1);

		if (EVTAG_GET(a, addr, &data, &data_len) == -1)
			return (-1);
		if (data_len != sizeof(addr))
			return (-1);
		memcpy(&addr, data, data_len);
		
		inf = TAILQ_FIRST(interfaces);

		add_request_to_interface(inf, &addr);

		/* shuffle around */
		TAILQ_REMOVE(interfaces, inf, next);
		TAILQ_INSERT_TAIL(interfaces, inf,next);

	}

	fprintf(stderr, " ... read %d probes\n", EVTAG_LEN(state, addresses));

	return (0);
}

int
resume_checkpoint(const char *filename,
    struct generators *generators,
    struct interfaceq *interfaces)
{
	extern uint32_t rndsbox[];
	struct evbuffer *buf = NULL;
	struct State *state = NULL;
	uint8_t *data;
	uint32_t data_len;
	int fd = -1; 

	fprintf(stderr, "... Reading from checkpoint \"%s\"\n", filename);

	fd = open(filename, O_RDONLY, 0);
	if (fd == -1) {
		warn("Cannot open \"%s\"\n", filename);
		return (-1);
	}

	buf = evbuffer_new();
	assert(buf != NULL);

	while (1) {
		int n = evbuffer_read(buf, fd, -1);
		if (n == 0)
			break;
		if (n == -1) {
			if (errno != EAGAIN)
				err(1, "failure while reading");
		}
	}

	close(fd); fd = -1;

	state = State_new();
	assert(state != NULL);

	if (evtag_unmarshal_State(buf, CHECKPOINT_TAG, state) == -1)
		goto error;

	evbuffer_free(buf); buf = NULL;

	if (read_generators(state, generators) == -1) 
		goto error;

	if (interfaces != NULL) {
		/*
		 * if we have a counter adjustment we don't need pending
		 * requests.
		 */
		if (read_requests(state, interfaces) == -1)
			goto error;
	}

	/* finally get the sbox state */
	if (EVTAG_GET(state, rndsbox, &data, &data_len) == -1)
		goto error;
	if (data_len != RNDSBOXSIZE * sizeof(uint32_t))
		goto error;
	memcpy((uint8_t *)rndsbox, data, data_len);

	State_free(state);

	return (0);

error:
	fprintf(stderr, " ... marshaled data has been corrupted\n");

	if (fd != -1)
		close(fd);
	if (buf != NULL)
		evbuffer_free(buf);
	if (state != NULL)
		State_free(state);

	return (-1);
}
