/*
 * DNSScan - simple DNS scanner
 *
 * Copyright 2007 (c) Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/queue.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <assert.h>

#include <event.h>
#include <evhttp.h>
#include <dnet.h>

#include "dnsscan.h"
#include "dnscore.h"
#include "exclude.h"
#include "utils.h"
#include "evutils.h"
#include "dnsdata.gen.h"
#include "checkpoint.h"
#include "state.h"
#include "fpdns.h"
#include "fpdns_states.h"

#define GENERATOR_DONE(x) ((x)->done)

#define MAKE_COUNTER(x, name) do { \
		struct timeval tvs[] = { \
			{  1, 0 }, \
			{ 60, 0 } \
		}; \
		(x) = event_counter_new(name, 2, tvs); \
		assert((x) != NULL); \
	} while (0)

/* proto types */
static void generators_adjust_bits(struct generators *generators);
static void open_result_file(const char *filename);
static int generate_address(struct addr *addr);
static void interface_enable_write(struct interface *inf);
static void readd_request_to_interface(
	struct interface *inf, struct request *req);

/* globals */
int use_http = 0;
int verbose = 0;

struct history_buffer *history;
static struct statetable *statetable;

struct event_histogram *hist_iquery;
struct event_histogram *hist_tc;

static struct interfaceq (interfaces);
static struct generators generators;

/* for http fingerprinting */
struct requestq http_requests;
int num_http_requests;
int num_pending_http_requests;

/* for distributing jobs */
static int num_servers = 1;
static int server_offset = 0;

/* rerun */
static int rerun_enabled;
static struct event ev_rerun;

/* results file */
int use_fsync = 0;
const char *result_file = "results.data";
static struct event ev_rotate;

/* address generation */
static float requests_per_second = 5.0;
static struct event ev_gen;
static struct timeval tv_gen;
static int result_fd = -1;
static int num_pending_requests;

static const char *checkpoint_filename = "checkpoint";
static struct event ev_checkpoint;

const char *query_suffix = ".osd.honeyd.org";
static int use_excludes = 1;
static int dont_send = 0;	/* true if we should not send packets */

static struct timeval tv_started;

static struct event_counter *ctr_sent;
static struct event_counter *ctr_received;
static struct event_counter *ctr_generated;
static struct event_counter *ctr_skipped;
struct event_counter *ctr_completed;
struct event_counter *ctr_resolver;
struct event_counter *ctr_recursive;
struct event_counter *ctr_chaos_answered;
struct event_counter *ctr_chaos_empty;
struct event_counter *ctr_web_server;
struct event_counter *ctr_skipped_web;

/* periodic rotating of result file */
static void
rotate_cb(int fd, short what, void *arg)
{
	struct timeval tv;

	if (result_fd != -1) {
		close(result_fd);
		result_fd = -1;
	}
		    
	timerclear(&tv);
	tv.tv_sec = DNS_ROTATE_INTERVAL_IN_SEC;
	event_add(&ev_rotate, &tv);
}

/* re-enable the generators */
static void
enable_generators_cb(int fd, short what, void *arg)
{
	struct timeval tv;

	if (!GENERATOR_DONE(&generators) || num_pending_requests)
		goto out;

	/* restart everything */
	generators.done = 0;
	generators.count = server_offset;
	generators_adjust_bits(&generators);

	/* also shuffle the seed */
	rndsboxinit(random());

out:	
	timerclear(&tv);
	tv.tv_sec = DNS_RERUN_INTERVAL_IN_SEC;
	event_add(&ev_rerun, &tv);
}

/* periodic checkpointing callback */

static void
checkpoint_cb(int fd, short what, void *arg)
{
	struct timeval tv;

	if (create_checkpoint(checkpoint_filename,
		&generators, &interfaces) != -1) {
		struct evbuffer *buf = evbuffer_new();
		assert(buf != NULL);
		evbuffer_add_printf(buf, "created successful checkpoint");
		history_enter_buffer(history, buf);
	}

	timerclear(&tv);
	tv.tv_sec = DNS_CHECKPOINT_INTERVAL_IN_SEC;
	event_add(&ev_checkpoint, &tv);
}

static void
signal_handler(int fd, short what, void *arg)
{
	fprintf(stderr, "... Terminating on received signal\n");

	create_checkpoint(checkpoint_filename, &generators, &interfaces);
	exit(1);
}

/* request generator */
static void
request_creator(int fd, short what, void *arg)
{
	static int generated;
	static int print_count;
	struct timeval tv, diff;
	int i, howmany;

	if (num_pending_requests >= DNS_MAX_PENDING_ADDRESSES)
		goto out;

	gettimeofday(&tv, NULL);
	timersub(&tv, &tv_gen, &diff);

	howmany = ((float)diff.tv_sec + (float)diff.tv_usec / 1000000.0) *
	    requests_per_second;

	/* don't print as often */
	if (++print_count == 100) {
		print_count = 0;
		if (generated) {
			fprintf(stderr, "Generated %d addresses\n", generated);
			generated = 0;
		}
	}

	if (howmany > 0) {
		/* do not generate too many at once */
		if (howmany > DNS_MAX_ADDRESSES_AT_ONCE)
			howmany = DNS_MAX_ADDRESSES_AT_ONCE;

		/*
		 * update the time only if we generated something, we make
		 * sure to keep the unused fractions right to be soo accurate
		 */
		diff.tv_sec = howmany / requests_per_second;
		diff.tv_usec = (howmany / requests_per_second - diff.tv_sec) * 
		    1000000L;
		timeradd(&tv_gen, &diff, &tv_gen);

		for (i = 0; i < howmany; ++i) {
			struct addr addr;
			struct interface *inf = TAILQ_FIRST(&interfaces);

			/* perhaps we are done? */
			if (generate_address(&addr) == -1) {
				/* keep track of how many we generated now */
				event_counter_add(ctr_generated, i);
				generated += i;

				/* we might not be able to generate any now */
				if (!GENERATOR_DONE(&generators) ||
				    rerun_enabled)
					goto out;

				/* we are completely done */
				return;
			}

			add_request_to_interface(inf, &addr);

			/* shuffle interfaces */
			TAILQ_REMOVE(&interfaces, inf, next);
			TAILQ_INSERT_TAIL(&interfaces, inf, next);
		}

		/* we got to generate all we wanted */
		event_counter_add(ctr_generated, howmany);
		generated += howmany;
	}
out:
	timerclear(&tv);
	tv.tv_usec = 10 * 1000L;	/* every 10 ms */
	event_add(&ev_gen, &tv);
}

/* sending the queries */
static int
send_query(struct interface *inf, struct request *req)
{
	static uint8_t buf[1024];
	struct sockaddr_in sa;
	struct state *state = req->state;
	int n, res;

	if (req->id == 0)
		req->id = random();

	if ((n = state->probe_init(state, req, buf, sizeof(buf))) == -1)
		return (-1);

	addr_ntos(&req->dest, (struct sockaddr *)&sa);
	sa.sin_port = htons(53);

	DFPRINTF(2, (stderr, "Sending %d bytes to %s\n",
		n, addr_ntoa(&req->dest)));

	res = dont_send ? 0 : sendto(inf->fd, buf, n, 0,
	    (struct sockaddr *)&sa, sizeof(sa));
	if (res > 0)
		event_counter_add(inf->ctr_bytes_sent, res);

	return (res);
}

static void
request_connection_free(int fd, short what, void *arg)
{
	struct evhttp_connection *evcon = arg;
	evhttp_connection_free(evcon);
}

void
request_free(struct request *req)
{
	event_del(&req->ev);

	if (req->payload)
		free(req->payload);
	if (req->chaos_payload)
		free(req->chaos_payload);
	if (req->chaos_reply)
		free(req->chaos_reply);

	if (req->fpdns_iquery)
		free(req->fpdns_iquery);
	if (req->fpdns_tc)
		free(req->fpdns_tc);

	if (req->fphttp_server)
		free(req->fphttp_server);
	if (req->fphttp_date)
		free(req->fphttp_date);
	if (req->fphttp_powered)
		free(req->fphttp_powered);
	if (req->http_body_payload)
		free(req->http_body_payload);
	if (req->ev_con) {
		/*
		 * XXX: bad hack to deal with the fact that we cannot free
		 * the connection from the call back.
		 */
		event_once(-1, EV_TIMEOUT,
		    request_connection_free, req->ev_con, NULL);
	}

	if (req->server_reply)
		HTTPServerReply_free(req->server_reply);

	free(req);
}

void
request_log(struct request *req)
{
	struct evbuffer *buf = evbuffer_new();
	struct DNSRequest *dr = DNSRequest_new();
	struct timeval tv;

	event_counter_add(ctr_completed, 1);

	/* we are finally done now - log everything */
	gettimeofday(&tv, NULL);
	EVTAG_ASSIGN(dr, tv_sec, tv.tv_sec);
	EVTAG_ASSIGN(dr, tv_usec, tv.tv_usec);

	EVTAG_ASSIGN(dr, remote_addr,
	    (uint8_t *)&req->dest, sizeof(req->dest));
	EVTAG_ASSIGN(dr, payload, req->payload, req->payload_size);

	/* allows us to differentiat between older versions */
	EVTAG_ASSIGN(dr, chaos_sent, 1);
	if (req->chaos_payload)
		EVTAG_ASSIGN(dr, chaos_payload,
		    req->chaos_payload, req->chaos_payload_size);
	if (req->chaos_reply)
		EVTAG_ASSIGN(dr, chaos_reply, req->chaos_reply);
	if (req->fpdns_iquery)
		EVTAG_ASSIGN(dr, fpdns_iquery, req->fpdns_iquery);
	if (req->fpdns_tc)
		EVTAG_ASSIGN(dr, fpdns_tc, req->fpdns_tc);

	if (req->fphttp_server)
		EVTAG_ASSIGN(dr, fphttp_server, req->fphttp_server);
	if (req->fphttp_date)
		EVTAG_ASSIGN(dr, fphttp_date, req->fphttp_date);
	if (req->fphttp_powered)
		EVTAG_ASSIGN(dr, fphttp_powered, req->fphttp_powered);
	if (req->fphttp_body_len)
		EVTAG_ASSIGN(dr, fphttp_body_len, req->fphttp_body_len);
	if (req->http_body_payload)
		EVTAG_ASSIGN(dr, http_body_payload,
		    req->http_body_payload, req->http_body_payload_size);
	if (req->server_reply)
		EVTAG_ASSIGN(dr, server_reply, req->server_reply);

	evtag_marshal_DNSRequest(buf, DNSSCAN_TAG_REQUEST, dr);

	DNSRequest_free(dr);

	/* open it only if we write data */
	if (result_fd == -1)
		open_result_file(result_file);

	while (EVBUFFER_LENGTH(buf) > 0) {
		if (evbuffer_write(buf, result_fd) == -1) {
			if (errno != EAGAIN)
				err(1, "writing buffer");
		}
	}

	evbuffer_free(buf);

	if (use_fsync)
		fsync(result_fd);
}

/* processes the reply and advances to the next state */

static void
interface_next_state(struct interface *inf, struct request *req,
    struct addr* addr, uint8_t *buf, size_t buf_len)
{
	struct state *state = req->state;
	int next_state = state->next_state(state, req, addr, buf, buf_len);
	/* if we stay in the same state, we take a time out */
	if (next_state == END_STATE) {
		(*statetable->finish_state)(req);
	} else if (next_state == ABORT_STATE) {
		(*statetable->abort_state)(req);
	} else {
		/* prepare for next state */
		req->retries = 0;
		req->id = 0;

		req->state = statetable_get(statetable, next_state);
		assert(req->state != NULL);
		readd_request_to_interface(inf, req);
	}
}

static void
request_timeout(int fd, short what, void *arg)
{
	struct request *req = arg;
	struct interface *inf = req->inf;

	TAILQ_REMOVE(&inf->inflight, req, next);
	inf->num_inflight--;

	if (++req->retries > DNS_MAX_REQUEST_RETRIES) {
		interface_next_state(inf, req, NULL, NULL, 0);
		return;
	}

	readd_request_to_interface(inf, req);
}

static void
interface_read(int fd, short what, void *arg)
{
	struct evbuffer *evbuf;
	static uint8_t buf[1024];
	struct interface *inf = arg;
	struct sockaddr sa;
	struct addr addr;
	struct request *request;
	uint16_t trans_id;

	socklen_t salen = sizeof(sa);
	int n = recvfrom(inf->fd, buf, sizeof(buf), 0, &sa, &salen);

	if (n == -1) {
		/* nothing really that we can do here */
		warn("recvfrom");
		return;
	}

	/* increase our counters */
	event_counter_add(inf->ctr_bytes_read, n);
	event_counter_add(ctr_received, 1);

	/* if we cannot even get the transid something is funky */
	if (dnscore_get_transid(buf, n, &trans_id) == -1)
		return;

	addr_ston(&sa, &addr);
	DFPRINTF(1, (stderr, "Received %d bytes from %s: id %d\n",
		n, addr_ntoa(&addr), trans_id));

	request = dnsscore_request_from_trans_id(&inf->inflight, trans_id);
	if (request == NULL) {
		evbuf = evbuffer_new();
		assert(evbuf != NULL);
		evbuffer_add_printf(evbuf,
		    "Received %d bytes from %s: id %d not matched\n",
		    n, addr_ntoa(&addr), trans_id);
		history_enter_buffer(history, evbuf);
		return;
	}

	if (addr_cmp(&addr, &request->dest) != 0) {
		/*
		 * some resolvers are really weird.  let's make sure, that
		 * they are at least in the same /24.  this assumes detailed
		 * knowledge of libdnet
		 */
		if (memcmp(&addr.addr_ip, &request->dest.addr_ip, 3) != 0) {
			evbuf = evbuffer_new();
			assert(evbuf != NULL);
			evbuffer_add_printf(evbuf,
			    "Got reply from %s for request sent to %s\n",
			    addr_ntoa(&addr), addr_ntoa(&request->dest));
			history_enter_buffer(history, evbuf);
			return;
		}
	}
	
	TAILQ_REMOVE(&inf->inflight, request, next);
	inf->num_inflight--;

	interface_next_state(inf, request, &addr, buf, n);
}

static void
interface_write(int fd, short what, void *arg)
{
	struct interface *inf = arg;
	struct timeval tv;
	struct request *req;
	int num_sent = 0;

	while ((req = TAILQ_FIRST(&inf->pending)) != NULL) {
		TAILQ_REMOVE(&inf->pending, req, next);
		--inf->num_pending;
		--num_pending_requests;

		if (send_query(inf, req) == -1) {
			warn("send: %s", addr_ntoa(&req->dest));
			if (errno != ENOBUFS &&
			    ++req->retries > DNS_MAX_REQUEST_RETRIES) {
				interface_next_state(inf, req, NULL, NULL, 0);
			} else {
				TAILQ_INSERT_TAIL(&inf->pending, req, next);
				++inf->num_pending;
				++num_pending_requests;
			}
			goto out;
		}

		/* move the request from pending to inflight */
		TAILQ_INSERT_TAIL(&inf->inflight, req, next);
		inf->num_inflight++;

		/* make sure that we get notified if the request times out */
		timerclear(&tv);
		tv.tv_sec = DNS_REQUEST_TIMEOUT_IN_SEC;
		event_add(&req->ev, &tv);

		if (++num_sent >= DNS_MAX_SEND_AT_ONCE)
			break;
	}

out:
	/* increase our counters */
	event_counter_add(ctr_sent, num_sent);

	if (TAILQ_FIRST(&inf->pending) != NULL)
		event_add(&inf->ev_write, NULL);
}

static void
add_interface(const char *name)
{
	struct interface *inf = calloc(1, sizeof(struct interface));
	assert(inf != NULL);

	inf->fd = make_socket(1, name, 16383 + (random() % 16384));
	if (inf->fd == -1)
		err(1, "make_socket: %s", name);

	inf->name = strdup(name);
	assert(inf->name != NULL);

	TAILQ_INSERT_TAIL(&interfaces, inf, next);

	TAILQ_INIT(&inf->pending);
	TAILQ_INIT(&inf->inflight);

	MAKE_COUNTER(inf->ctr_bytes_sent, "Bytes Sent");
	MAKE_COUNTER(inf->ctr_bytes_read, "Bytes Read");

	event_set(&inf->ev_write, inf->fd, EV_WRITE, interface_write, inf);

	event_set(&inf->ev_read, inf->fd, EV_READ|EV_PERSIST,
	    interface_read, inf);
	event_add(&inf->ev_read, NULL);

	fprintf(stderr, "... Adding interface for %s\n", name);
}

static void
generators_adjust_bits(struct generators *generators)
{
	uint32_t last_count = 0, old_last_count = 0;
	struct generate *gen;
	int i = 0, entries = 0;
	uint32_t remaining = -1;
	
	/* compute a fast search index */
	TAILQ_FOREACH(gen, &generators->generators, next) {
		++entries;
	}

	if (generators->index != NULL)
		free(generators->index);

	generators->index = malloc(entries * sizeof(struct generate_index));
	assert(generators->index != NULL);
	generators->num_index = entries;

	TAILQ_FOREACH(gen, &generators->generators, next) {
		struct generate_index *idx = &generators->index[i++];
		if (TAILQ_FIRST(&generators->generators) != gen) {
			++last_count;
			if (!remaining)
				goto error;
			remaining -= 1;
		}

		if (remaining < gen->last_count) 
			goto error;
		remaining -= gen->last_count;

		/* this is the first count of the new generator */
		idx->first_count = last_count;
		idx->gen = gen;

		last_count += gen->last_count;
		/* avoid overflow by making sure that the count increases */
		assert(old_last_count <= last_count);
		old_last_count = last_count;
	}

	generators->last_real_count = last_count;

	/* recompute the total number of bits we need */
	if (last_count != -1) {
		generators->bits = 0;
		while (last_count) {
			++generators->bits;
			last_count >>= 1;
		}
		generators->bits = 32 - generators->bits;
	} else {
		generators->bits = 0;
	}

	if (generators->bits)
		generators->last_count = (1 << (32 - generators->bits)) - 1;
	else
		generators->last_count = -1;
	return;

error:
	errx(1, "Too many generators; sum larger than /0");

}

static void
add_generator(const struct addr *addr)
{
	struct generate *gen;
	struct addr net;

	gen = calloc(1, sizeof(struct generate));
	assert(gen != NULL);

	addr_net(addr, &net);
	gen->prefix = net.addr_ip;
	gen->bits = addr->addr_bits;

	if (gen->bits)
		gen->last_count = (1 << (32 - gen->bits)) - 1;
	else
		gen->last_count = -1;
	assert(gen->bits != 0 || gen->last_count == -1);

	TAILQ_INSERT_TAIL(&generators.generators, gen, next);
}

static int
generate_address_from_count(struct addr *addr, uint32_t count)
{
	extern struct address_list excludequeue;
	extern struct address_list rndexclqueue;
	struct generate *gen;

	int left = 0;
	int right = generators.num_index;

	/* binary search for the right spot */
	while (1) {
		int off = (right - left) / 2 + left;
		struct generate_index *idx = &generators.index[off];

		if (count < idx->first_count) {
			right = off;
		} else if (count <= idx->first_count + idx->gen->last_count) {
			gen = idx->gen;
			count -= idx->first_count;
			break;
		} else {
			left = off + 1;
		}
	}

	/* create the to be scanned address */
	addr->addr_type = ADDR_TYPE_IP;
	addr->addr_bits = 32;
	addr->addr_ip = gen->prefix | htonl(count);

	if (!use_excludes) {
		/* 
		 * not using any exclude lists; so all addresses
		 * are good.
		 */
		return (1);
	}

	if (list_included(*addr, &rndexclqueue))
		return (0);

	if (list_included(*addr, &excludequeue))
		return (0);

	return (1);
}

static int
generate_address(struct addr *addr)
{
	int attempts = DNS_MAX_GENERATE_ATTEMPTS;
	int skipped = 0;
	int found = 0;
	uint32_t left;

	if (GENERATOR_DONE(&generators))
		return (-1);

	do {
		uint32_t count = \
		    rndgetaddr(generators.bits, generators.count);

		if (count <= generators.last_real_count) {
			found = generate_address_from_count(addr, count);
		}
		
		if (!found)
			skipped++;

		/* move on to the next address */
		if (generators.count == generators.last_count) {
			generators.done = 1;
			break;
		}

		left =  generators.last_count - generators.count;
		if (left < num_servers) {
			generators.done = 1;
			break;
		}
		generators.count += num_servers;
	} while (!found && --attempts > 0);

	/* remember how many addressed we skipped */
	event_counter_add(ctr_skipped, skipped);

	return (found == 0 ? -1 : 0);
}

/* request for the interface */

static void
interface_enable_write(struct interface *inf)
{
	if (!event_pending(&inf->ev_write, EV_WRITE, NULL))
		event_add(&inf->ev_write, NULL);
}

void
add_request_to_interface(struct interface *inf, struct addr *addr)
{
	struct request *req = calloc(1, sizeof(struct request));

	req->inf = inf;
	req->dest = *addr;
	req->state = statetable_get(statetable, statetable->initial_state);
	assert(req->state != NULL);

	/* event_add when needed */
	evtimer_set(&req->ev, request_timeout, req);

	TAILQ_INSERT_TAIL(&inf->pending, req, next);
	++inf->num_pending;
	++num_pending_requests;

	interface_enable_write(inf);
}

static void
readd_request_to_interface(struct interface *inf, struct request *req)
{
	/* delete the timeout */
	event_del(&req->ev);

	TAILQ_INSERT_HEAD(&inf->pending, req, next);
	++inf->num_pending;
	++num_pending_requests;

	/* enable write only if it is not enabled already */
	if (TAILQ_FIRST(&inf->pending) == req)
		interface_enable_write(inf);
}

static void
quit_server(struct evhttp_request *request, void *arg)
{
	struct timeval tv;
	
	timerclear(&tv);
	tv.tv_sec = 5;
	fprintf(stderr, "... quiting in %ld seconds\n", tv.tv_sec);
	event_loopexit(&tv);

	evhttp_send_reply(request, HTTP_OK, "OK", NULL);
}

/* display our current status */

static char css_style[] = 
    ".container { "
    "  position: relative; "
    "}"
    ".eta { "
    "  position: absolute; "
    "  top: 0px; "
    "  left: 520px; "
    "  font-family: Verdana, Arial; "
    "  font-size: 10pt; "
    "}"
    ".hundred { "
    "  position: relative; "
    "  background-color: #ee8888; "
    "  border: solid 1px; "
    "  margin: 0.5em; "
    "  width: 500px; "
    "  height: 15px; "
    "}"
    ".progress { "
    "  position: absolute; "
    "  top: 0px; "
    "  left: 0px; "
    "  background-color: #88ee88; "
    "  margin: 0em; "
    "  height: 15px; "
    "}"
    ".percent { "
    "  position: absolute; "
    "  top: 0px; "
    "  left: 0px; "
    "  width: 500px; "
    "  font-family: Verdana, Arial; "
    "  font-weight: bold;"
    "  font-size: 10pt; "
    "  text-align: center; "
    "}"
    ".tiny {\n"
    "  color: #bbbbcc;\n"
    "  padding: 2px 0px 0px 2px;"
    "  margin-bottom: -10em;"
    "  font-size: 0.5em;\n"
    "  font-family: Verdana, Arial;\n"
    "}\n"
    "table.generator {"
    "  margin-right: 1em;"
    "  width: 400px;"
    "  border-width: 0px 0px 1px 1px;"
    "  border-spacing: 2px;"
    "  border-style: inset;"
    "  border-color: black;"
    "  border-collapse: collapse;"
    "}"
    "table.generator th {"
    "  border-width: 1px 1px 0px 0px;"
    "  padding: 2px;"
    "  border-style: inset;"
    "  border-color: black;"
    "  background-color: rgb(255, 240, 240);"
    "  font-family: Verdana, Arial;\n"
    "  font-size: small;\n"
    "}\n"
    "table.generator td {"
    "  border-width: 1px 1px 0px 0px;"
    "  padding: 2px;"
    "  border-style: inset;"
    "  border-color: black;"
    "  background-color: rgb(255, 250, 220);"
    "  font-family: Verdana, Arial;\n"
    "  font-size: small;\n"
    "}\n"
    ".banner {\n"
    "  background: #ddddff;\n"
    "  font-family: Verdana, Arial;\n"
    "  border: 1px solid;\n"
    "}\n"
    ".banner h1 {\n"
    "  margin-top: -0.3em;\n"
    "  margin-bottom: 0.1em;\n"
    "  text-align: center;\n"
    "}\n"
    ".stats {\n"
    "  padding: 5px;\n"
    "  font-family: Verdana, Arial;\n"
    "  padding: 0.25em;\n"
    "  font-size: 0.9em;\n"
    "}\n"
    ".generators {\n"
    "  padding: 5px;\n"
    "  clear: both; \n"
    "}\n"
    "\n"
    ".generators h2 {\n"
    "  border: 1px solid;\n"
    "  background-color: #ddccff;\n"
    "  font-family: Verdana, Arial;\n"
    "  padding: 0.25em;\n"
    "  font-size: 1.25em;\n"
    "}\n"
    ".generators h3 {"
    "  border: 1px solid;"
    "  margin-right: 1em;"
    "  padding: 2px; "
    "  background-color: #fff0f0;"
    "  font-family: Verdana, Arial;"
    "  font-size: 1em;"
    "  text-align: center;"
    "}"
    "\n"
    ".inlined {\n"
    "  padding: 0px;\n"
    "  margin: 0px;\n"
    "  display: inline;\n"
    "}\n"
    "\n"
    ".events {\n"
    "  padding: 5px;\n"
    "}\n"
    "\n"
    ".events h2 {\n"
    "  border: 1px solid;\n"
    "  background-color: #ddccff;\n"
    "  font-family: Verdana, Arial;\n"
    "  padding: 0.25em;\n"
    "  font-size: 1.25em;\n"
    "}";

static void
css_server(struct evhttp_request *request, void *arg)
{
	struct evbuffer *databuf = evbuffer_new();
	assert(databuf != NULL);
	evhttp_add_header(request->output_headers, "Content-Type", "text/css");
	evbuffer_add(databuf, css_style, strlen(css_style));

	/* send along our data */
	evhttp_send_reply(request, HTTP_OK, "OK", databuf);
	evbuffer_free(databuf);
}

static void
show_event_count(struct evbuffer *databuf, struct event_counter *cnt)
{
	uint32_t counts[3];

	event_counter_get(cnt, counts);

	/* skip counters that do not contain any content */
	if (!counts[0] && !counts[1] && !counts[2])
		return;

	evbuffer_add_printf(databuf,
	    "<tr><td>%s</td><td>%d</td><td>%d</td><td>%d</td></tr>\n",
	    cnt->name, counts[0], counts[1], counts[2]);

}

static void
show_generator_status(struct evbuffer *databuf)
{
	static char eta[50];
	struct generate *gen = NULL;
	float complete;
	int count = 0;

	double tmp;
	time_t seconds_remaining = 0;

	complete = 1.0;
	if (!GENERATOR_DONE(&generators)) {
		if (generators.last_count) {
			complete = (float)generators.count/
			    (float)generators.last_count;
		} else {
			complete = (float)generators.count; /* 0 or 1 */
		}

		/* figure out how much longer we need */
		tmp = (double)(generators.last_count - generators.count) /
		    (double)(requests_per_second * num_servers);
		if (ctr_generated->total) {
			/* adjust for rate with which we generate addresses */
			tmp *= (double)(ctr_generated->total)/
			    (double)(ctr_generated->total+ctr_skipped->total);
		}
		seconds_remaining = tmp;
	}

	event_seconds_as_eta(seconds_remaining, eta, sizeof(eta));

	evbuffer_add_printf(databuf,
	    "<div class=\"container\">"
	    "<div class=\"hundred\">"
	    "<div class=\"progress\" style=\"width: %dpx\">&nbsp;</div>"
	    "<div class=\"percent\">%.1f%%</div>"
	    "</div>"
	    "<div class=\"eta\">"
	    "ETA <b>%s</b> (%u/%u)"
	    "</div>"
	    "</div>\n",
	    (int)(complete * 500.0), complete * 100.0,
	    eta,
	    generators.count, generators.last_count
	    );

	evbuffer_add_printf(databuf, "<ul>\n");
	TAILQ_FOREACH(gen, &generators.generators, next) {
		struct addr addr;
		/* only display some but not everything */
		if (++count > 1000) {
			evbuffer_add_printf(databuf,
			    "<li class=\"inlined\">...</li>\n");
			break;
		}

		addr.addr_type = ADDR_TYPE_IP;
		addr.addr_bits = gen->bits;
		addr.addr_ip = gen->prefix;

		evbuffer_add_printf(databuf, "<li class=\"inlined\">%s</li>\n",
		    addr_ntoa(&addr));
	}
	evbuffer_add_printf(databuf, "</ul>\n");
}

static void
show_statistics(struct evbuffer *databuf)
{
	evbuffer_add_printf(databuf, "<p style=\"clear: both\"/>");

	evbuffer_add_printf(databuf, "<div style=\"float: left\">");
	evbuffer_add_printf(databuf, "<h3>General</h3>\n");
	evbuffer_add_printf(databuf, "<table class=generator>\n");
	evbuffer_add_printf(databuf,
	    "<tr><th>Type</th><th>Minute</th>"
	    "<th>Hour</th><th>Total</th></tr>\n");
	show_event_count(databuf, ctr_sent);
	show_event_count(databuf, ctr_received);
	show_event_count(databuf, ctr_generated);
	show_event_count(databuf, ctr_skipped);
	show_event_count(databuf, ctr_completed);
	show_event_count(databuf, ctr_resolver);
	show_event_count(databuf, ctr_recursive);
	show_event_count(databuf, ctr_chaos_answered);
	show_event_count(databuf, ctr_chaos_empty);
	show_event_count(databuf, ctr_web_server);
	show_event_count(databuf, ctr_skipped_web);
	evbuffer_add_printf(databuf, "</table>\n");
	evbuffer_add_printf(databuf, "</div>");

	evbuffer_add_printf(databuf, "<div style=\"float: left\">");
	evbuffer_add_printf(databuf, "<h3>Fingerprint IQUERY</h3>\n");
	event_histogram_as_html(databuf, "generator", hist_iquery);
	evbuffer_add_printf(databuf, "</div>");

	evbuffer_add_printf(databuf, "<div style=\"float: left\">");
	evbuffer_add_printf(databuf, "<h3>Fingerprint TC</h3>\n");
	event_histogram_as_html(databuf, "generator", hist_tc);
	evbuffer_add_printf(databuf, "</div>");

	evbuffer_add_printf(databuf, "<p style=\"clear: both\"/>");
}

static void
show_interface_status(struct evbuffer *databuf)
{
	struct interface *inf;
	uint32_t bytes_read[3], bytes_sent[3];

	evbuffer_add_printf(databuf, "<ul>\n");
	evbuffer_add_printf(databuf,
	    "<li>general: dns: pending <b>%d</b>, "
	    "http: pending <b>%d</b>, inflight <b>%d</b></li>\n",
	    num_pending_requests,
	    num_pending_http_requests, num_http_requests);
	TAILQ_FOREACH(inf, &interfaces, next) {
		double read_kbps, sent_kbps;
		event_counter_get(inf->ctr_bytes_read, bytes_read);
		read_kbps = (double)bytes_read[0] / 60.0 * 8.0 / 1024.0;
		event_counter_get(inf->ctr_bytes_sent, bytes_sent);
		sent_kbps = (double)bytes_sent[0] / 60.0 * 8.0 / 1024.0;

		evbuffer_add_printf(databuf,
		    "<li>"
		    "<b>%s</b>: pending <b>%u</b>, inflight <b>%u</b>, "
		    "sent: <b>%.1f</b> kbps, read: <b>%.1f</b> kbps"
		    "</li>",
		    inf->name, inf->num_pending, inf->num_inflight,
		    sent_kbps, read_kbps);
	}
	evbuffer_add_printf(databuf, "</ul>\n");
}

static void
show_status_cb(struct evhttp_request *req, void *arg)
{
	static char running[50];
	struct timeval tv;
	struct evbuffer *databuf = evbuffer_new();
	assert(databuf != NULL);

	evbuffer_add_printf(databuf,
	    "<html><head>\n"
	    "<link rel=stylesheet type=text/css href=/css>\n"
	    "<title>DNSScan Status</title></head>\n");
	evbuffer_add_printf(databuf,
	    "<body>\n"
	    "<div class=banner>"
	    "<span class=tiny>(c) 2007 Niels Provos</span> "
	    "<h1>DNSScan Status</h1>\n"
	    "</div>\n");

	gettimeofday(&tv, NULL);
	timersub(&tv, &tv_started, &tv);
	event_seconds_as_eta(tv.tv_sec, running, sizeof(running));
	evbuffer_add_printf(databuf, 
	    "<div class=stats>"
	    "Running for: %s %s\n",
	    running, rerun_enabled ? "continous scan" : "");

	if (num_servers > 1) {
		evbuffer_add_printf(databuf, "<br />"
		    "Distributed Scan %d/%d\n",
		    server_offset, num_servers);
	}

	evbuffer_add_printf(databuf, "</div>");

	evbuffer_add_printf(databuf,
	    "<div class=generators><h2>Interfaces</h2>");
	show_interface_status(databuf);
	evbuffer_add_printf(databuf, "</div>\n");

	evbuffer_add_printf(databuf,
	    "<div class=generators><h2>Statistics</h2>");
	show_statistics(databuf);
	evbuffer_add_printf(databuf, "</div>\n");

	evbuffer_add_printf(databuf,
	    "<div class=generators><h2>Generators</h2>");
	show_generator_status(databuf);
	evbuffer_add_printf(databuf, "</div>\n");

	evbuffer_add_printf(databuf,
	    "<div class=events><h2>Events</h2>");
	history_to_html(databuf, history);
	evbuffer_add_printf(databuf, "</div>\n");

	evbuffer_add_printf(databuf,
	    "</body></html>");

	/* send along our data */
	evhttp_send_reply(req, HTTP_OK, "OK", databuf);
	evbuffer_free(databuf);
}

static void
open_result_file(const char *filename)
{
	static char destination[1024];
	struct timeval tv;
	struct evbuffer *buf = evbuffer_new();
	assert(buf != NULL);

	gettimeofday(&tv, NULL);

	snprintf(destination, sizeof(destination), "%s.%ld",
	    filename, tv.tv_sec);

	/* open up results file */
	result_fd = open(destination, O_WRONLY|O_APPEND|O_CREAT, 0600);
	if (result_fd == -1)
		err(1, "failure opening %s", destination);

	evbuffer_add_printf(buf, "writing results to %s", destination);
	history_enter_buffer(history, buf);
}

static void
parse_address_entry(const char *entry)
{
	/* do something */
	struct addr addr, net;
	if (addr_pton(entry, &addr) == -1)
		errx(1, "cannot parse %s", entry);
	/* write the network part out */
	addr_net(&addr, &net);
	net.addr_bits = addr.addr_bits;

	add_generator(&net);
}

static struct option dnsscan_long_opts[] = {
	{ "seed", required_argument, NULL, 'r' },
	{ "checkpoint", required_argument, NULL, 'c' },
	{ "disable_sending", no_argument, &dont_send, 1},
	{ "use_http", no_argument, &use_http, 1 },
	{ "verbose", 0, NULL, 'v' },
	{ "result_file", required_argument, NULL, 'w' },
	{ "num_servers", required_argument, NULL, 'N' },
	{ "server_offset", required_argument, NULL, 'O' },
	{ "detection_file", required_argument, NULL, 'D' },
	{ "query_suffix", required_argument, NULL, 'Q' },
	{ 0, 0, 0, 0}
};
	

static void
usage(const char *name)
{
	fprintf(stderr, 
	    "%s: [-e excludefile] [-i if] <IP address|network>...\n\n"
	    "\t-d	   do not send any network packets\n"
	    "\t-c <file>   resume from checkpoint\n"
	    "\t-r <seed>   seed for random number generator,\n"
	    "\t-V          print version number of scanssh,\n"
	    "\t-e <file>   exclude the IP addresses and networks in <file>,\n"
	    "\t-p <port>   port for web server,\n"
	    "\t-I <ipaddr> address of interface to bind web server to,\n"
	    "\t-v          increase verbosity level"
	    "\t-i <alias>  specifies the IP alias to connect from,\n"
	    "\t-R	   rerun all addresses continously\n"
	    "\t-A <offset> adjust the current count by the offset\n",
	    name);
}

void
setup_reporting(void)
{
	MAKE_COUNTER(ctr_sent, "Sent Packets");
	MAKE_COUNTER(ctr_received, "Received Packets");
	MAKE_COUNTER(ctr_generated, "Addresses Generated");
	MAKE_COUNTER(ctr_skipped, "Addresses Skipped");
	MAKE_COUNTER(ctr_completed, "Completed Requests");
	MAKE_COUNTER(ctr_resolver, "Open Resolvers");
	MAKE_COUNTER(ctr_recursive, "Open Recursive");
	MAKE_COUNTER(ctr_chaos_answered, "Chaos Answered");
	MAKE_COUNTER(ctr_chaos_empty, "Chaos Empty");
	MAKE_COUNTER(ctr_web_server, "Web Servers");
	MAKE_COUNTER(ctr_skipped_web, "Skipped Web Fetch");

	history = history_buffer_new(100);
	assert(history != NULL);

	hist_iquery = event_histogram_new(50);
	assert(hist_iquery);

	hist_tc = event_histogram_new(50);
	assert(hist_tc);
}

int
main(int argc, char **argv)
{
	struct event ev_sigint, ev_sigterm;
	struct evhttp *http = NULL;
	const char *name = argv[0];
	extern char *excludefile;
	struct timeval tv;
	const char *webserver_interface = "127.0.0.1";
	uint16_t port = 8080;
	int32_t adjust_counter = 0;
	const char *resume_filename = NULL;
	const char *generate_filename = NULL;
	int seed = 123456;
	int ch;

	TAILQ_INIT(&interfaces);

	TAILQ_INIT(&http_requests);

	memset(&generators, 0, sizeof(generators));
	TAILQ_INIT(&generators.generators);

	event_init();
	evtag_init();

	while ((ch = getopt_long(argc, argv,
		    "vdXVN:f:w:p:c:r:I:i:e:q:RA:Q:",
		    dnsscan_long_opts, NULL)) != -1)
		switch(ch) {
		case 0:
			/* everything is good */
			break;
		case 'Q':
			if (optarg[0] != '.')
				errx(1, "Bad query suffix: %s", optarg);
			query_suffix = optarg;
			break;
		case 'N':
			num_servers = atoi(optarg);
			if (num_servers <= 0)
				errx(1, "Bad number of servers: %s", optarg);
			break;
		case 'O':
			server_offset = atoi(optarg);
			if (server_offset < 0)
				errx(1, "Bad server offset: %s", optarg);
			break;
		case 'A':
			adjust_counter = atol(optarg);
			if (!adjust_counter)
				errx(1, "Bad counter adjustment: %s", optarg);
			break;
		case 'v':
			++verbose;
			break;
		case 'p':
			port = atoi(optarg);
			if (port == 0)
				errx(1, "Bad port number: %s", optarg);
			break;
		case 'I':
			webserver_interface = optarg;
			break;
		case 'd':
			fprintf(stderr, "... disabling sending\n");
			dont_send = 1;
			break;
		case 'c':
			resume_filename = optarg;
			break;
		case 'w':
			result_file = optarg;
			break;
		case 'f':
			generate_filename = optarg;
			break;
		case 'r':
			seed = atoi(optarg);
			break;
		case 'i':
			add_interface(optarg);
			break;
		case 'q':
			requests_per_second = atof(optarg);
			if (!requests_per_second)
				errx(1, "Illegal qps value: %s", optarg);
			break;
		case 'X':
			use_excludes = 0;
			break;
		case 'V':
			fprintf(stderr, "DNSScan %s\n", VERSION);
			exit(0);
		case 'e':
			excludefile = optarg;
			break;
		case 'R':
			rerun_enabled = 1;
			break;
		case 'h':
		default:
			usage(name);
			exit(1);
		}

	if (server_offset >= num_servers)
		errx(1, "Bad server offset %d >= %d\n",
		    server_offset, num_servers);

	argc -= optind;
	argv += optind;

	if (TAILQ_FIRST(&interfaces) == 0)
		add_interface("0.0.0.0");

	setupexcludes();
	rndsboxinit(seed);

	fprintf(stderr, "... running in FPDNS mode\n");

	statetable = setup_fpdns_states();

	/* set up result file rotation */
	evtimer_set(&ev_rotate, rotate_cb, NULL);
	timerclear(&tv);
	tv.tv_sec = DNS_ROTATE_INTERVAL_IN_SEC;
	event_add(&ev_rotate, &tv);

	/* set up rerunning of generators */
	evtimer_set(&ev_rerun, enable_generators_cb, NULL);
	if (rerun_enabled) {
		timerclear(&tv);
		tv.tv_sec = DNS_RERUN_INTERVAL_IN_SEC;
		event_add(&ev_rerun, &tv);
	}

	if (!argc && generate_filename == NULL) {
		if (resume_filename == NULL) {
			fprintf(stderr, "Nothing to be done\n");
			exit(0);
		}

		if (resume_checkpoint(resume_filename,
			&generators,
			!adjust_counter ? &interfaces : NULL) == -1) {
			fprintf(stderr, "Failed to resume checkpoint\n");
			exit(1);
		}
	} else if (server_offset) {
		fprintf(stderr, "... setting offset to %d\n", server_offset);
		generators.count = server_offset;
	}


	/* start requests as soon as they are available */
	gettimeofday(&tv_gen, NULL);
	
	evtimer_set(&ev_gen, request_creator, NULL);

	timerclear(&tv);
	event_add(&ev_gen, &tv);

	/* set up check pointing */
	evtimer_set(&ev_checkpoint, checkpoint_cb, NULL);
	timerclear(&tv);
	tv.tv_sec = DNS_CHECKPOINT_INTERVAL_IN_SEC;
	event_add(&ev_checkpoint, &tv);

	while (argc) {
		parse_address_entry(argv[0]);

		argv++;
		argc--;
	}

	if (generate_filename) {
		int count = 0;
		int fd = open(generate_filename, O_RDONLY, 0);
		if (fd == -1)
			err(1, "Cannot open file \"%s\"", generate_filename);
		struct evbuffer *buf = evbuffer_new();
		assert(buf != NULL);
		
		while (evbuffer_read(buf, fd, -1) > 0) {
			char *line;
			while ((line = evbuffer_readline(buf)) != NULL) {
				++count;
				parse_address_entry(line);
				free(line);
			}
		}

		fprintf(stderr, "... read %d generators\n", count);

		evbuffer_free(buf);
		close(fd);
	}

	/* make sure that the random algorithm works */
	generators_adjust_bits(&generators);

	/* sometimes we need to adjust our internal count forward/backward */
	if (adjust_counter) {
		generators.count = generators.count + adjust_counter;
		fprintf(stderr, "... adjusted %d\n", generators.count);
	}

	/* ignore SIGPIPE */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
		err(1, "signal(SIGPIPE)");

	/* intercept sigint and sigterm */
	signal_set(&ev_sigint, SIGINT, signal_handler, NULL);
	signal_add(&ev_sigint, NULL);
	signal_set(&ev_sigterm, SIGTERM, signal_handler, NULL);
	signal_add(&ev_sigterm, NULL);

	/* set up our visual status reporting */

	setup_reporting();

	http = evhttp_start(webserver_interface, port);
	if (http == NULL)
		errx(1, "Cannot start HTTP server");
	evhttp_set_cb(http, "/", show_status_cb, http);
	evhttp_set_cb(http, "/css", css_server, http);
	evhttp_set_cb(http, "/_quit", quit_server, http);

	fprintf(stderr, "... Started HTTP server at %s:%d\n",
	    webserver_interface, port);

	gettimeofday(&tv_started, NULL);

	event_dispatch();

	return (0);
}
