/*
 * Copyright 2007 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <assert.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "dnscore.h"

#define APPEND8(x) do {				\
		if (j + 1 > (off_t)buf_len)	\
			goto overflow;		\
		_t8 = x;			\
		buf[j] = _t8;			\
		j += 1;				\
	} while (0)
#define APPEND16(x) do {			\
		if (j + 2 > (off_t)buf_len)	\
			goto overflow;		\
		_t16 = htons(x);		\
		memcpy(buf + j, &_t16, 2);	\
		j += 2;				\
	} while (0)
#define APPEND32(x) do {			\
		if (j + 4 > (off_t)buf_len)	\
			goto overflow;		\
		_t32 = htonl(x);		\
		memcpy(buf + j, &_t32, 4);	\
		j += 4;				\
	} while (0)

int
fpdns_probe_iquery(uint8_t *buf, size_t buf_len,
    uint16_t trans_id, const char *name)
{
	uint8_t _t8; uint16_t _t16; uint32_t _t32;
	off_t len_idx, name_start, r, len_name;

	int j = 0;

	APPEND16(trans_id);
	APPEND16(0x1000);	/* iquery, no other flags */
	APPEND16(0);		/* zero questions */
	APPEND16(1);		/* one answer */
	APPEND16(0);		/* no authority */
	APPEND16(0);		/* no additional */

	APPEND8(0);		/* zero name */
	APPEND16(0);		/* type */
	APPEND16(0);		/* class */
	APPEND32(0);		/* ttl */

	/* append the name */
	len_idx = j;
	j += 2;
	name_start = j;
	r = dnscore_name_to_labels(buf, buf_len, j, name, strlen(name), NULL);
	if (r < 0)
		goto overflow;
	
	len_name = j - name_start;
	j = len_idx;
	APPEND16(len_name);
	j = r;

	return (j);

overflow:
	return (-1);
}

int
fpdns_probe_tc(uint8_t *buf, size_t buf_len,
    uint16_t trans_id, const char *name)
{
	uint8_t _t8; uint16_t _t16; uint32_t _t32;
	int j = 0;

	APPEND16(trans_id);
	APPEND16(0x0200);	/* query, TC-bit set */
	APPEND16(1);		/* zero questions */
	APPEND16(0);		/* one answer */
	APPEND16(0);		/* no authority */
	APPEND16(0);		/* no additional */

	APPEND8(0);		/* zero name */
	APPEND16(0);		/* type */
	APPEND16(0);		/* class */
	APPEND32(0);		/* ttl */

	j = dnscore_name_to_labels(buf, buf_len, j, name, strlen(name), NULL);
	if (j < 0)
		goto overflow;
	
	APPEND16(EVDNS_TYPE_A);
	APPEND16(EVDNS_CLASS_INET);

	return (j);

overflow:
	return (-1);
}

/* interpreting answers */

#define GET32(x) do { if (j + 4 > length) goto err; memcpy(&_t32, packet + j, 4); j += 4; x = ntohl(_t32); } while(0)
#define GET16(x) do { if (j + 2 > length) goto err; memcpy(&_t, packet + j, 2); j += 2; x = ntohs(_t); } while(0)
#define GET8(x) do { if (j >= length) goto err; x = packet[j++]; } while(0)

#define QR(x)		(((x) & 0x8000) != 0)
#define OPCODE(x)	(((x) & 0x3800) >> 8)
#define AA(x)		(((x) & 0x0400) != 0)
#define TC(x)		(((x) & 0x0200) != 0)
#define RD(x)		(((x) & 0x0100) != 0)
#define RA(x)		(((x) & 0x0080) != 0)
#define Z(x)		(((x) & 0x0040) != 0)
#define AD(x)		(((x) & 0x0020) != 0)
#define CD(x)		(((x) & 0x0010) != 0)
#define RCODE(x)	((x) & 0x000f)	

#define ADDFP(x) do {				\
		if (k + strlen(x) >= out_len)	\
			goto err;		\
		memcpy(out + k, (x), strlen(x));	\
		k += strlen(x);			\
	} while (0)


int
fpdns_print(uint8_t *packet, off_t length, char *out, off_t out_len)
{
	int j = 0, k = 0;
	uint16_t _t;

	uint16_t trans_id, questions, answers, authority;
	uint16_t additional;
        uint16_t flags = 0;

	if (packet == NULL) {
		ADDFP("t");
		goto out;
	}
	
	GET16(trans_id);
	GET16(flags);
	GET16(questions);
	GET16(answers);
	GET16(authority);
	GET16(additional);

	/* d. j. bernstein classification */
	switch (RCODE(flags)) {
	case 0:
		ADDFP("0");
		break;
	case 1:
		ADDFP("1");
		break;
	case 2:
		ADDFP("2");
		break;
	case 3:
		ADDFP("3");
		break;
	case 4:
		ADDFP("4");
		break;
	case 5:
		ADDFP("5");
		break;
	}

	if (TC(flags))
		ADDFP("TC");
	if (RD(flags))
		ADDFP("RD");
	if (AA(flags))
		ADDFP("AA");
	if (Z(flags))
		ADDFP("Z1");
	if (AD(flags))
		ADDFP("Z2");
	if (CD(flags))
		ADDFP("Z3");

	if (!questions)
		ADDFP("q");
	else if (questions == 2)
		ADDFP("Q2");
	if (answers)
		ADDFP("D");

out:
	/* add terminating character */
	if (k >= out_len)
		goto err;
	out[k] = '\0';

	return (0);

err:
	return (-1);
}

int
fpdns_test(void)
{
	static uint8_t buf[128];
	static uint8_t answers_iprobe[] = {
		0x3c, 0xf9, 0x90, 0x01, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00
	};
	static char ascii_fp[20];

	int r = fpdns_probe_tc(buf, sizeof(buf), 65535, "osd.honeyd.org");

	assert(r > 0);

	fpdns_print(answers_iprobe, sizeof(answers_iprobe),
	    ascii_fp, sizeof(ascii_fp));
	fprintf(stderr, "FP: %s\n", ascii_fp);

	exit(0);
}
